#ifndef PEBBLISP_PLFUNC_H
#define PEBBLISP_PLFUNC_H

#include "pebblisp.h"

#define BASIC_OP(_name) \
    Object _name(Object* params, int length, struct Environment* env)

BASIC_OP(add);

BASIC_OP(sub);

BASIC_OP(mul);

BASIC_OP(dvi);

BASIC_OP(mod);

BASIC_OP(equ);

BASIC_OP(gth);

BASIC_OP(lth);

BASIC_OP(and);

BASIC_OP(or);

#undef BASIC_OP

/// ANY => STRING
fn(catObjects,
   "Concatenate string versions of the given objects.",
   "(cat \"Stuff: \" (1 2 3))", "Stuff: ( 1 2 3 )",
);

/// FUNCY, LIST => LIST
fn(filter,
   "Filter a list based on the given condition.",
   "(fil (fn (a) (< 50 a)) (25 60 100))", "( 60 100 )",
   "(fil (fn (a) (< 0 (len a))) ( () (1) (1 2) () ))", "( ( 1 ) ( 1 2 ) )",
);

/// LIST, ANY => LIST
fn(append,
   "Append the given element. Creates a new list.",
   "(ap (1 2) 3)", "( 1 2 3 )",
);

/// LIST, ANY => LIST
fn(prepend,
   "Prepend the given element. Creates a new list",
   "(pre (2 3) 1)", "( 1 2 3 )",
);

tfn(len,
    ({ TYPE_LIST, TYPE_NUMBER }),
    "Returns the length of the given list, or a NOT_A_LIST error if the expression is not a list.",
    "(len (2 3))", "2",
    "(len ())", "0",
    "(len \"string\")", "NOT_A_LIST",
);

/// LIST, FUNCY, ANY => ANY
fn(reduce,
   "Performs a simple reduction. Does not currently work with lambdas.\n"
   "Takes three arguments:\n"
   " - Values\n"
   " - A function to apply to each value\n"
   " - An initial value",
   "(reduce 5 + 6)", "11",
   "(reduce (1 2 3) + 0)", "6",
);

/// NUMBER, LIST => ANY
fn(at,
   "Get item at the given index in the given list.",
   "(at 1 (1 2 3))", "2",
   "(at 99 (1 2 3))", "INDEX_PAST_END",
   "(at 99 \"string\")", "INDEX_PAST_END",
);

/// LIST => LIST
fn(rest,
   "Get the tail of a list. All but the first element.",
   "(rest (1 2 3))", "( 2 3 )",
   "(rest ())", "( )",
   "(rest \"string\")", "NOT_A_LIST",
);

/// LIST => LIST
fn(reverse,
   "Reverse a list.",
   "(rev (1 2 3))", "( 3 2 1 )",
   "(rev \"string\")", "NOT_A_LIST",
);

/// ANY => BOOL
fn(isNum,
   "Returns `T` only if the argument evaluates to a number.",
   "(isnum 1)", "T",
   "(isnum (+ 5 5))", "T",
   "(isnum '(+ 5 5))", "F",
   "(isnum \"Hello\")", "F",
);

/// ANY => BOOL
fn(isList,
   "Returns `T` only if the argument is a list.",
   "(islist (1 2 3))", "T",
   "(islist ())", "T",
   "(islist \"Stringy\")", "F",
);

/// ANY => BOOL
fn(isString,
   "Returns `T` only if the argument is a string.",
   "(isstr \"Heyo\")", "T",
   "(isstr \"\")", "T",
   "(isstr (cat 5 5))", "T",
   "(isstr 10)", "F",
);

/// ANY => BOOL
fn(isErr,
   "Check if the argument is an error.",
   "(iserr (at 10 ()))", "T",
   "(iserr 5)", "F",
);

/// STRING => STRING
fn(charAt,
   "Get the char in the given string at the given index.",
   "(chat \"Hello\" 1)", "e",
   "(chat \"Hello\" 10)", "",
);

/// STRING => NUMBER
fn(charVal,
   "Get the ascii integer representaton of the given character.",
   "(char \"h\")", "104",
   "(char \"hello\")", "104",
   "(char \"\")", "0",
);

/// STRING/SLIST => ANY
fn(parseEvalO,
   "Evaluate the given string or quoted list.",
   "(eval \"(1 2 3)\")", "( 1 2 3 )",
   "(eval '(+ 5 5))", "10",
);

/// STRUCT, STRING => ANY
fn(possessive,
   "(struct Post (title body))\n"
   "(def p (Post \"TI\" \"BO\"))\n"
   "p's title => TI"
);

#ifdef STANDALONE

fn(print, "Prints the string representation of the given object to stdout.");

fn(numToChar,
   "Gets a string containing the ascii character for the given number value.",
   "(ch 107)", "k",
   "(ch 0x21)", "!",
);

fn(printEnvO, "Prints out the current scoped environment.");

fn(systemCall,
   "Opens a shell and runs the given command, returning 0 if successful.\n"
   "If the argument is not a string, returns 255.\n",
   "(sys \"echo yee > /dev/null\")", "0",
   "(sys 5)", "255",
);

fn(loadFile,
   "Loads and parses the given file.\n"
   "Returns 0 if the file was loaded and parsed successfully. Otherwise 1.\n"
   "(loadfile \"printdate.pl\")\n"
   "Mon 21 Mar 2022 10:35:03 AM EDT\n"
   "=> 0"
);

/// @code
/// () => STRING
/// STRING => STRING
fn(takeInput,
   "Take console input with an optional prompt. For example:\n"
   "`(def x (input))` will wait for user input with no prompt.\n"
   "`(def x (input \">> \"))` wait for input, but prompt the user with '>> '.\n"
);

/// STRING => STRING
fn(help,
   "Displays help text for the given function.\n"
   "Currently requires the function name as a string, but future syntactic sugar may\n"
   "loosen this requirement.\n"
   "(? \"+\") => \"(+ 1 2) => 3\""
);

/// STRING => STRING
fn(readFileToObject,
   "Read a file into a string object."
);

#endif // STANDALONE

#endif // PEBBLISP_PLFUNC_H